# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Map and globe widgets for Jupyter Notebooks using Remote Frame Buffer."""

# Dependencies: jupyter_rfb, pillow, imageio, simplejpeg, ipycanvas

__all__ = ['GlobeWidget', 'MapWidget', 'GfxAnalysisWidget']

import asyncio
from ctypes import CFUNCTYPE, Structure, addressof, byref, c_int, c_size_t, c_void_p, cast, cdll, pointer
import os
import time

from IPython.display import display
from jupyter_rfb import RemoteFrameBuffer
from jupyter_rfb._png import array2png
import numpy as np

from ...internal.comutil import (
    CLSCTX_INPROC_SERVER,
    E_NOINTERFACE,
    GUID,
    HRESULT,
    LPVOID,
    POINTER,
    PVOID,
    REFIID,
    S_OK,
    ULONG,
    IUnknown,
    OLE32Lib,
    Succeeded,
)
from ...internal.stkxrfb import IRemoteFrameBuffer, IRemoteFrameBufferHost
from ...stkobjects import IAnimation, Scenario, STKObjectRoot
from ...stkx import ButtonValues, Graphics2DControlBase, Graphics3DControlBase, GraphicsAnalysisControlBase, ShiftValues
from ...utilities.exceptions import STKAttributeError

TIMERPROC = CFUNCTYPE(None, c_size_t)
INSTALLTIMER = CFUNCTYPE(c_size_t, c_int, TIMERPROC, c_void_p)
DELETETIMER = CFUNCTYPE(c_int, c_size_t, c_void_p)


class AsyncioTimerManager(object):
    """Provide timer support for animation in jupyter notebooks."""
    class TimerInfo(object):
        def __init__(self, id, milliseconds, timer_proc, callback_data):
            """Construct an object of type TimerInfo."""
            self.id = id
            self.interval = milliseconds/1000
            self.callback = timer_proc
            self.callback_data = callback_data
            self._reset()

        def _reset(self):
            self.next_proc = time.perf_counter() + self.interval

        def fire(self):
            if time.perf_counter() >= self.next_proc:
                self.callback(self.id)
                self._reset()

    def __init__(self):
        """Construct an object of type AsyncioTimerManager."""
        if os.name != 'nt':
            agutillib = cdll.LoadLibrary("libagutil.so")
        else:
            agutillib = cdll.LoadLibrary("AgUtil.dll")

        functype = CFUNCTYPE(None, INSTALLTIMER, DELETETIMER, c_void_p)
        set_timer_callbacks = functype(
                ("AgUtSetTimerCallbacks", agutillib),
                (
                    (1, "pInstallTimer"),
                    (1, "pDeleteTimer"),
                    (1, "pCallbackData")
                )
        )

        self._next_id = 1
        self._timers = dict()
        self._install_timer_cfunc = INSTALLTIMER(self.__install_timer)
        self._delete_timer_cfunc = DELETETIMER(self.__delete_timer)
        set_timer_callbacks(
            self._install_timer_cfunc,
            self._delete_timer_cfunc, c_void_p())

    def terminate(self):
        self._timers.clear()

    def __install_timer(self, milliseconds, timer_proc, callback_data):
        id = self._next_id
        self._next_id = id + 1
        self._timers[id] = AsyncioTimerManager.TimerInfo(
            id, milliseconds, timer_proc, callback_data)
        self._set_alarm_for_next_timer_proc()
        return id

    def __delete_timer(self, timer_id, callback_data):
        if timer_id in self._timers:
            del(self._timers[timer_id])
        return 0

    def _fire_timers(self):
        timers = self._timers.copy()
        for timerid in timers:
            timers[timerid].fire()
        self._set_alarm_for_next_timer_proc()

    async def _wait(self, delay):
        await asyncio.sleep(delay)
        self._fire_timers()

    def _next_timer_proc(self):
        """Return time in sec until next timer proc."""
        temp_timers = self._timers.copy()
        if len(temp_timers) == 0:
            return 0.050
        else:
            proc_times = list()
            for timerid in temp_timers:
                proc_times.append(temp_timers[timerid].next_proc)
            delta_s = min(proc_times) - time.perf_counter()
            if delta_s > 0:
                return delta_s
            else:
                return 0

    def _set_alarm_for_next_timer_proc(self):
        next_proc = self._next_timer_proc()
        if next_proc > 0:
            self.task = asyncio.ensure_future(self._wait(next_proc))
        else:
            self._fire_timers()


asyncio_timer_manager = None


class RemoteFrameBufferHostVTable(Structure):
    """Structure of the vtable for IRemoteFrameBufferHost."""
    _fields_ = [("IUnknown1",        c_void_p),
                ("IUnknown2",        c_void_p),
                ("IUnknown3",        c_void_p),
                ("refresh",          c_void_p)]


class RemoteFrameBufferHost(object):
    """
    Implements IRemoteFrameBufferHost.

    Assemble a vtable following the layout of that interface
    """
    _iid_unknown = GUID(IUnknown._guid)
    _iid_iagremoteframebufferhost = GUID('{D229A605-D3A8-4476-B628-AC549C674B58}')

    def __init__(self, owner):
        """Construct an object of type RemoteFrameBufferHost."""
        self.owner = owner
        self._init_vtable()

    def _init_vtable(self):

        qi = CFUNCTYPE(HRESULT, PVOID, REFIID, POINTER(PVOID))(self._query_interface)
        addref = CFUNCTYPE(ULONG, PVOID)(self._add_ref)
        release = CFUNCTYPE(ULONG, PVOID)(self._release)

        if os.name == "nt":
            self.__dict__['_cfunc_IUnknown1'] = qi
            self.__dict__['_cfunc_IUnknown2'] = addref
            self.__dict__['_cfunc_IUnknown3'] = release
        else:
            self.__dict__['_cfunc_IUnknown3'] = qi
            self.__dict__['_cfunc_IUnknown1'] = addref
            self.__dict__['_cfunc_IUnknown2'] = release

        self.__dict__['_cfunc_Refresh'] = CFUNCTYPE(None, PVOID)(self._refresh)

        self.__dict__['_vtable'] = RemoteFrameBufferHostVTable(
            *[cast(self._cfunc_IUnknown1, c_void_p),
              cast(self._cfunc_IUnknown2, c_void_p),
              cast(self._cfunc_IUnknown3, c_void_p),
              cast(self._cfunc_Refresh, c_void_p)]
        )
        self.__dict__['_unknown'] = pointer(self._vtable)

    def _add_ref(self, this: PVOID) -> int:
        return 1

    def _release(self, this: PVOID) -> int:
        return 0

    def _query_interface(self,
                        this: PVOID,
                        riid: REFIID,
                        object: POINTER(PVOID)) -> int:
        iid = riid.contents
        if iid == RemoteFrameBufferHost._iid_unknown:
            object[0] = addressof(self._unknown)
            return S_OK
        elif iid == RemoteFrameBufferHost._iid_iagremoteframebufferhost:
            object[0] = addressof(self._unknown)
            return S_OK
        else:
            object[0] = 0
            return E_NOINTERFACE

    def _refresh(self, this: PVOID) -> None:
        self.owner.request_draw()


class WidgetBase(RemoteFrameBuffer):
    """Base class for Jupyter controls."""
    _shift = 0x0001
    _control = 0x0004
    _lalt = 0x0008
    _ralt = 0x0080
    _mouse1 = 0x0100
    _mouse2 = 0x0200
    _mouse3 = 0x0400

    def __init__(self,
                 root: STKObjectRoot,
                 w: int = 800,
                 h: int = 600,
                 title: str = None,
                 resizable: bool = True):
        """Construct an object of type WidgetBase."""
        super().__init__()

        self.frame = None

        self.css_width = f"{w}px"
        self.css_height = f"{h}px"

        self.resizable = resizable
        self.pixel_ratio = 1.0

        self._unk = self.__create_instance(self._progid)

        self._interface._private_init(self, self._unk)

        self.__create_frame_buffer(w, h)

        self._rfb = IRemoteFrameBuffer(self)
        self._rfb.set_to_offscreen_rendering(w, h)

        self._rfbHostImpl = RemoteFrameBufferHost(self)

        self._rfbHostImplUnk = IUnknown()
        self._rfbHostImplUnk.p = addressof(self._rfbHostImpl._unknown)

        self._rfbHost = IRemoteFrameBufferHost()
        self._rfbHost._private_init(self._rfbHostImplUnk)

        self._rfb.set_host(self._rfbHost)

        self._building_examples = os.getenv("BUILD_EXAMPLES", "false") == "true"

        self.mouse_callbacks = [
            [
                self._rfb.notify_left_button_down,
                self._rfb.notify_right_button_down,
                self._rfb.notify_middle_button_down
            ],
            [
                self._rfb.notify_left_button_up,
                self._rfb.notify_right_button_up,
                self._rfb.notify_middle_button_up
            ]
        ]

        global asyncio_timer_manager
        if asyncio_timer_manager is None:
            asyncio_timer_manager = AsyncioTimerManager()

        self.root = root
        if self.root.current_scenario is not None:
            self.title = title or self.root.current_scenario.instance_name
            scenario: Scenario = Scenario(self.root.current_scenario)
            if scenario.scene_manager.scenes.count > 0:
                self.camera = scenario.scene_manager.scenes.item(0).camera

    def __del__(self):
        del self._rfb
        del self._rfbHostImpl
        del self._rfbHost
        del self._unk
        self.root = None
        self.camera = None

    def __create_frame_buffer(self, w: int, h: int):
        if self.frame is not None:
            del self.frame
        self.frame = np.ones((int(h), int(w), 3), np.uint8)
        self.pointer, read_only_flag = self.frame.__array_interface__['data']

    def __create_instance(self, clsid: str) -> LPVOID:
        guid = GUID()
        if Succeeded(OLE32Lib.CLSIDFromString(clsid, guid)):
            iid_iunknown = GUID(IUnknown._guid)
            unk = IUnknown()
            if Succeeded(OLE32Lib.CoCreateInstance(byref(guid), None,
                                          CLSCTX_INPROC_SERVER,
                                          byref(iid_iunknown), byref(unk.p))):
                unk.take_ownership()
                return unk
        return None

    def __setattr__(self, attrname, value):
        try:
            self._interface.__setattr__(self, attrname, value)
        except STKAttributeError:
            RemoteFrameBuffer.__setattr__(self, attrname, value)

    def __get_modifiers(self, event):
        modifiers = event['modifiers']
        result = 0
        if "Shift" in modifiers:
            result = result | ShiftValues.PRESSED
        if "Ctrl" in modifiers:
            result = result | ShiftValues.CTRL_PRESSED
        if "Alt" in modifiers:
            result = result | ShiftValues.ALT_PRESSED
        return result

    def __get_position(self, event):
        x = int(event["x"] * self.pixel_ratio)
        y = int(event["y"] * self.pixel_ratio)
        return (x, y)

    def handle_event(self, event):

        event_type = event.get("event_type", None)
        if event_type == "resize":
            pixel_ratio = event["pixel_ratio"]
            w = int(event["width"]*pixel_ratio)
            h = int(event["height"]*pixel_ratio)
            self.pixel_ratio = pixel_ratio
            self.__create_frame_buffer(w, h)
            self._rfb.notify_resize(0, 0, w, h)
        elif event_type == "pointer_down":
            (x, y) = self.__get_position(event)
            self.mouse_callbacks[0][event["button"]-1](
                x, y, self.__get_modifiers(event))
        elif event_type == "pointer_up" and event["button"] == 1:
            (x, y) = self.__get_position(event)
            self.mouse_callbacks[1][event["button"]-1](
                x, y, self.__get_modifiers(event))
        elif event_type == "pointer_move":
            (x, y) = self.__get_position(event)
            buttons = event["buttons"]
            if len(buttons) > 0 and buttons[0] == 1:
                self._rfb.notify_mouse_move(x, y, ButtonValues.LEFT_PRESSED,
                                          self.__get_modifiers(event))
            elif len(buttons) > 0 and buttons[0] == 2:
                self._rfb.notify_mouse_move(x, y, ButtonValues.RIGHT_PRESSED,
                                          self.__get_modifiers(event))
            elif len(buttons) > 0 and buttons[0] == 3:
                self._rfb.notify_mouse_move(x, y, ButtonValues.MIDDLE_PRESSED,
                                          self.__get_modifiers(event))
            else:
                self._rfb.notify_mouse_move(x, y, 0, 0)
        elif event_type == "wheel":
            (x, y) = self.__get_position(event)
            dy = int(event["dy"] * self.pixel_ratio/100)
            self._rfb.notify_mouse_wheel(x, y, -dy, self.__get_modifiers(event))

    def set_title(self, title):
        self.title = title

    def get_frame(self):
        self._rfb.snap_to_rbg_raster(self.pointer)
        return self.frame

    def animate(self, time_step):
        scenario: Scenario = Scenario(self.root.current_scenario)
        scenario.animation_settings.animation_step_value = time_step
        animation_control: IAnimation = IAnimation(self.root)
        animation_control.play_forward()
        self.show()

    def _repr_mimebundle_(self, **kwargs):
        """Return the desired MIME type representation.

        The MIME type representation is a dictionary relating MIME types to
        the data that should be rendered in that format.

        The main goal of this function is to provide the right type of data when
        renedring different types of documents, including HTML, Notebooks, and
        PDF files.

        """
        if not self._building_examples:
            data = super()._repr_mimebundle_(**kwargs)
        else:
            data = {
                "image/png": array2png(self.snapshot().data)
            }
        return data

    def show(self, in_sidecar=False, **snapshot_kwargs):
        if in_sidecar:
            from sidecar import Sidecar
            with Sidecar(title=self.title):
                display(self)
        else:
            return self

    def snapshot(self, pixel_ratio=None, _initial=False):
        if self._building_examples:
            # There is currently no good way to detect when terrain and
            # imagery fetched in background threads are finished loading,
            # so when building the documentation to capture snapshots,
            # just use an arbitrary delay for now and fetch a few frames
            # before taking the snapshot
            for _ in range(0, 4):
                _ = self.get_frame()
                time.sleep(0.5)
        return super().snapshot(pixel_ratio, _initial)

class GlobeWidget(Graphics3DControlBase, WidgetBase):
    """The 3D Globe widget for jupyter."""
    # Example:
    #   from ansys.stk.core.stkengine import *
    #   from ansys.stk.core.stkengine.jupyterwidgets import GlobeWidget

    #   stk = STKEngine.StartApplication(noGraphics=False)
    #   root = stk.NewObjectRoot()
    #   g = GlobeWidget(root, 600, 400)
    #   root.NewScenario("RemoteFrameBuffer")
    #   root.ExecuteCommand('Animate * Start Loop')
    #   g

    _progid = "STKX12.VOControl.1"
    _interface = Graphics3DControlBase

    def __init__(self, root: STKObjectRoot, w: int, h: int, title: str = None):
        """Construct an object of type GlobeWidget."""
        WidgetBase.__init__(self, root, w, h, title)

    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        WidgetBase.__setattr__(self, attrname, value)


class MapWidget(Graphics2DControlBase, WidgetBase):
    """The 2D Map widget for jupyter."""
    _progid = "STKX12.2DControl.1"
    _interface = Graphics2DControlBase

    def __init__(self, root: STKObjectRoot, w: int, h: int, title: str = None):
        """Construct an object of type MapWidget."""
        WidgetBase.__init__(self, root, w, h, title)

    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        WidgetBase.__setattr__(self, attrname, value)


class GfxAnalysisWidget(GraphicsAnalysisControlBase, WidgetBase):
    """The Graphics Analysis widget for jupyter."""
    _progid = "STKX12.GfxAnalysisControl.1"
    _interface = GraphicsAnalysisControlBase

    def __init__(self, root: STKObjectRoot, w: int, h: int, title: str = None):
        """Construct an object of type GfxAnalysisWidget."""
        WidgetBase.__init__(self, root, w, h, title)

    def __setattr__(self, attrname, value):
        """Attempt to assign an attribute."""
        WidgetBase.__setattr__(self, attrname, value)



# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Starts STK Desktop or attaches to an already running STK Desktop, and provides access to the Object Model root."""

__all__ = ["STKDesktop", "STKDesktopApplication"]

import atexit
from ctypes import byref
import os
import pathlib
import socket

# The subprocess module is needed to start the backend.
# Excluding low severity bandit warning as the validity of the inputs is enforced.
import subprocess  # nosec B404
import typing

from .internal.apiutil import InterfaceProxy, read_registry_key, winreg_stk_binary_dir
from .internal.coclassutil import attach_to_stk_by_pid
from .internal.comutil import (
    CLSCTX_LOCAL_SERVER,
    CO_E_NOTINITIALIZED,
    COINIT_APARTMENTTHREADED,
    GUID,
    PVOID,
    CoInitializeManager,
    IUnknown,
    ObjectLifetimeManager,
    OLE32Lib,
    OLEAut32Lib,
    Succeeded,
)
from .internal.eventutil import EventSubscriptionManager
from .stkobjects import STKObjectModelContext, STKObjectRoot
from .uiapplication import UiApplication
from .utilities.exceptions import STKInitializationError, STKRuntimeError
from .utilities.grpcutilities import (
    GrpcAuthenticationMode,
    GrpcCallBatcher,
    _get_authentication_mode_string,
    _validate_authentication_mode,
)


class ThreadMarshaller(object):
    """Automate multiple STK instances from one Python script using threads."""
    _iid_iunknown = GUID.from_registry_format(IUnknown._guid)
    def __init__(self, obj):
        if os.name != "nt":
            raise STKRuntimeError("ThreadMarshaller is only available on Windows.")
        if not hasattr(obj, "_intf"):
            raise STKRuntimeError("Invalid object to passed to ThreadMarshaller.")
        if type(obj._intf) is not IUnknown:
            raise STKRuntimeError("ThreadMarshaller is not available on the gRPC API.")
        self._obj = obj
        self._obj_type = type(obj)
        self._pStream = PVOID()
        if not Succeeded(OLE32Lib.CoMarshalInterThreadInterfaceInStream(byref(ThreadMarshaller._iid_iunknown), obj._intf.p, byref(self._pStream))):
            raise STKRuntimeError("ThreadMarshaller failed to initialize.")

    def __del__(self):
        if self._pStream is not None:
            OLE32Lib.CoReleaseMarshalData(self._pStream)
        del(self._obj)

    def get_marshalled_to_current_thread(self) -> typing.Any:
        """Return an instance of the original stk_object that may be used on the current thread. May only be called once."""
        if self._pStream is None:
            raise STKRuntimeError(f"{self._obj_type} object has already been marshalled to a thread.")
        unknown_raw = PVOID()
        hr = OLE32Lib.CoGetInterfaceAndReleaseStream(self._pStream, byref(ThreadMarshaller._iid_iunknown), byref(unknown_raw))
        self._pStream = None
        if not Succeeded(hr):
            if hr == CO_E_NOTINITIALIZED:
                raise STKRuntimeError("Thread not initialized. Call InitializeThread() before the call to GetMarshalledToCurrentThread().")
            else:
                raise STKRuntimeError("Could not marshall to thread.")
        unknown = IUnknown()
        unknown.p = unknown_raw
        marshalled_obj = self._obj_type()
        marshalled_obj._private_init(unknown)
        del(unknown)
        return marshalled_obj

    def initialize_thread(self) -> None:
        """Must be called on the destination thread prior to calling GetMarshalledToCurrentThread()."""
        OLE32Lib.CoInitializeEx(None, COINIT_APARTMENTTHREADED)

    def release_thread(self) -> None:
        """Call in the destination thread after all calls to STK are finished."""
        OLE32Lib.CoUninitialize()

class STKDesktopApplication(UiApplication):
    """
    Interact with an STK Desktop application.

    Use STKDesktop.StartApplication() or STKDesktop.AttachToApplication()
    to obtain an initialized STKDesktopApplication object.
    """

    def __init__(self):
        """Construct an object of type STKDesktopApplication."""
        if os.name != "nt":
            raise STKRuntimeError("STKDesktopApplication is only available on Windows. Use STKEngine.")
        self.__dict__["_intf"] = InterfaceProxy()
        UiApplication.__init__(self)
        self.__dict__["_root"] = None

    def _private_init(self, intf: InterfaceProxy):
        UiApplication._private_init(self, intf)

    def __del__(self):
        """Destruct the STKDesktopApplication object after all references to the object are deleted."""
        if self._intf and type(self._intf) is IUnknown:
            CoInitializeManager.uninitialize()

    @property
    def root(self) -> STKObjectRoot:
        """Get the object model root associated with this instance of STK Desktop application."""
        if not self._intf:
            raise STKRuntimeError("STKDesktopApplication has not been properly initialized.  Use STKDesktop to obtain the STKDesktopApplication object.")
        if self._root is not None:
            return self._root
        if self._intf:
            self.__dict__["_root"] = self.personality2
            return self.__dict__["_root"]

    def new_object_model_context(self) -> STKObjectModelContext:
        """Create a new object model context for the STK Desktop application."""
        return self.create_object("{7A12879C-5018-4433-8415-5DB250AFBAF9}", "")

    def set_grpc_options(self, options:dict) -> None:
        """
        Set advanced-usage options for the gRPC client.

        Available options include:
        { "collection iteration batch size" : int }. Number of items to preload while iterating
        through a collection object. Default is 100. Use 0 to indicate no limit (load entire collection).
        { "disable batching" : bool }. Disable all batching operations.
        { "release batch size" : int }. Number of interfaces to be garbage collected before
        sending the entire batch to STK to be released. Default value is 12.
        """
        if hasattr(self._intf, "client"):
            self._intf.client.set_grpc_options(options)

    def new_grpc_call_batcher(self, max_batch:int=None, disable_batching:bool=False) -> GrpcCallBatcher:
        """
        Construct a GrpcCallBatcher linked to this gRPC client that may be used to improve API performance.

        If gRPC is not active, the batcher will be disabled.
        max_batch is the maximum number of calls to batch together.
        Set disable_batching=True to disable batching operations for this batcher.
        See grpcutilities module for more information.
        """
        if hasattr(self._intf, "client"):
            batcher = GrpcCallBatcher(disable_batching)
            batcher._private_init(self._intf.client, max_batch)
        else:
            batcher = GrpcCallBatcher(disable_batching=True)
        return batcher

    def shutdown(self) -> None:
        """Close this STK Desktop instance (or detach if the instance was obtained through STKDesktop.AttachToApplication())."""
        if self._root is not None:
            root : STKObjectRoot = self._root
            root.close_scenario()
            self.__dict__["_root"] = None
        if hasattr(self._intf, "client"):
            self.user_control = False
            self._disconnect_grpc()
        else:
            self.quit()
            self.__dict__["_intf"] = InterfaceProxy()

    def _disconnect_grpc(self) -> None:
        """Safely disconnect from STK."""
        if self._intf:
            if typing.TYPE_CHECKING:
                from .internal.grpcutil import GrpcClient
            client: GrpcClient = self._intf.client
            client.terminate_connection()
            self.__dict__["_intf"] = InterfaceProxy()


class STKDesktop(object):
    """Create, initialize, and manage STK Desktop application instances."""

    _disable_pop_ups = False

    @staticmethod
    def start_application(visible:bool=False,
                         user_control:bool=False,
                         grpc_server:bool=False,
                         grpc_host:str="127.0.0.1",
                         grpc_port:int=40704,
                         grpc_timeout_sec:int=60,
                         grpc_max_message_size:int=0,
                         grpc_allow_remote_host:bool=False,
                         grpc_server_cert:str=None,
                         grpc_server_key:str=None,
                         grpc_client_cert:str=None,
                         grpc_client_key:str=None,
                         grpc_ca:str=None,
                         grpc_authentication_mode:GrpcAuthenticationMode=GrpcAuthenticationMode.SINGLE_USER) -> STKDesktopApplication:
        """
        Create a new STK Desktop application instance.

        Specify visible = True to show the application window.
        Specify user_control = True to return the application to the user's control.
        (the application remains open) after terminating the Python API connection.
        Specify grpc_server = True to attach to STK Desktop Application running the gRPC server at grpc_host:grpc_port.
        grpc_host is the IP address or DNS name of the gRPC server.
        grpc_port is the integral port number that the gRPC server is using (valid values are integers from 0 to 65535).
        grpc_timeout_sec specifies the time allocated to wait for a grpc connection (seconds).
        grpc_max_message_size is the maximum size in bytes that the gRPC client can receive. Set to zero to use the gRPC default.
        Specify grpc_allow_remote_host = True to allow external connections, not allowed by default. Required when using 0.0.0.0 or other remote address as the grpc_host.
        grpc_server_cert is the path to the server certificate file. Required for mTLS authentication.
        grpc_server_key is the path to the server key file. Required for mTLS authentication.
        grpc_client_cert is the path to the client certificate file. Required for mTLS authentication.
        grpc_client_key is the path to the client key file. Required for mTLS authentication.
        grpc_ca is the path to the issuing certificate authority. Required for mTLS authentication.
        grpc_authentication_mode is the method of client-server authentication to use for gRPC. Default is SINGLE_USER.
        Only available on Windows.
        """
        if os.name != "nt":
            raise STKRuntimeError("STKDesktop is only available on Windows. Use STKEngine.")

        CoInitializeManager.initialize()
        if grpc_server:
            try:
                pass
            except ModuleNotFoundError:
                raise STKInitializationError("gRPC use requires Python modules grpcio and protobuf.")
            if grpc_port < 0 or grpc_port > 65535:
                raise STKInitializationError(f"{grpc_port} is not a valid port number for the gRPC server.")
            if grpc_host not in ["localhost", "127.0.0.1"]:
                if not grpc_allow_remote_host:
                    raise RuntimeError("Remote host connections are not allowed. Use grpc_allow_remote_host to enable.")
                try:
                    socket.inet_pton(socket.AF_INET, grpc_host)
                except OSError:
                    try:
                        socket.inet_pton(socket.AF_INET6, grpc_host)
                    except OSError:
                        raise STKInitializationError(f"Could not resolve host \"{grpc_host}\" for the gRPC server.")

            _validate_authentication_mode(grpc_authentication_mode, grpc_host)

            clsid_stk12application = "{7ADA6C22-FA34-4578-8BE8-65405A55EE15}"
            executable = read_registry_key(f"CLSID\\{clsid_stk12application}\\LocalServer32", silent_exception=True)
            if executable is None or not pathlib.Path(executable).exists():
                bin_dir = pathlib.Path(winreg_stk_binary_dir()).resolve()
                if bin_dir.exists():
                    executable = bin_dir / "AgUiApplication.exe"
                else:
                    raise STKInitializationError("Could not find AgUiApplication.exe. Verify STK 12 installation.")
            cmd_line = [f"{executable}", "/pers", "STK", "/grpcServer", "On", "/grpcHost", grpc_host, "/grpcPort", str(grpc_port)]

            cmd_line.extend(["/grpcAuthMode", _get_authentication_mode_string(grpc_authentication_mode)])
            if grpc_server_cert:
                cmd_line.extend(["/grpcServerCert", str(pathlib.Path(grpc_server_cert).resolve())])
            if grpc_server_key:
                cmd_line.extend(["/grpcServerKey", str(pathlib.Path(grpc_server_key).resolve())])
            if grpc_ca:
                cmd_line.extend(["/grpcCa", str(pathlib.Path(grpc_ca).resolve())])
            if grpc_allow_remote_host:
                cmd_line.append("/grpcAllowRemoteHost")
            if STKDesktop._disable_pop_ups:
                cmd_line.append("/Automation")

            # Calling subprocess.Popen (without shell equals true) to start the backend.
            # Excluding low severity bandit check as the validity of the inputs has been ensured.
            p = subprocess.Popen(cmd_line) # nosec B603
            host = grpc_host
            # Ignoring B104 warning as it is a false positive. The hard-coded string "0.0.0.0" is being filtered
            # to ensure that it is not used.
            if grpc_host=="0.0.0.0": # nosec B104
                host = "127.0.0.1"
            try:
                app = STKDesktop.attach_to_application(None,
                                                     grpc_server,
                                                     host,
                                                     grpc_port,
                                                     grpc_timeout_sec,
                                                     grpc_max_message_size,
                                                     grpc_allow_remote_host,
                                                     grpc_client_cert,
                                                     grpc_client_key,
                                                     grpc_ca,
                                                     grpc_authentication_mode)

                app.visible = visible
                app.user_control = user_control
                return app
            except Exception:
                if not user_control:
                    p.terminate()
                raise
        else:
            clsid_aguiapplication = GUID()
            if Succeeded(OLE32Lib.CLSIDFromString("STK12.Application", clsid_aguiapplication)):
                unknown = IUnknown()
                iid_iunknown = GUID(IUnknown._guid)
                if Succeeded(OLE32Lib.CoCreateInstance(byref(clsid_aguiapplication), None, CLSCTX_LOCAL_SERVER, byref(iid_iunknown), byref(unknown.p))):
                    unknown.take_ownership(isApplication=True)
                    app = STKDesktopApplication()
                    app._private_init(unknown)
                    app.visible = visible
                    app.user_control = user_control
                    return app
            raise STKInitializationError("Failed to create STK Desktop application.  Check for successful install and registration.")

    @staticmethod
    def attach_to_application(pid:int=None,
                            grpc_server:bool=False,
                            grpc_host:str="127.0.0.1",
                            grpc_port:int=40704,
                            grpc_timeout_sec:int=60,
                            grpc_max_message_size:int=0,
                            grpc_allow_remote_host:bool=False,
                            grpc_client_cert:str=None,
                            grpc_client_key:str=None,
                            grpc_ca:str=None,
                            grpc_authentication_mode:GrpcAuthenticationMode=GrpcAuthenticationMode.SINGLE_USER) -> STKDesktopApplication:
        """
        Attach to an existing STK Desktop instance.

        Specify the Process ID (PID) in case multiple processes are open.
        Specify grpc_server = True to attach to STK Desktop Application running the gRPC server at grpc_host:grpc_port.
        grpc_host is the IP address or DNS name of the gRPC server.
        grpc_port is the integral port number that the gRPC server is using.
        grpc_timeout_sec specifies the time allocated to wait for a grpc connection (seconds).
        grpc_max_message_size is the maximum size in bytes that the gRPC client can receive. Set to zero to use the gRPC default.
        Specify grpc_allow_remote_host = True to allow external connections, not allowed by default. Required when using 0.0.0.0 or other remote address as the grpc_host.
        grpc_client_cert is the path to the client certificate. Required to perform mTLS authentication.
        grpc_client_key is the path to the client private encryption key. Required to perform mTLS authentication.
        grpc_ca is the path to the issuing certificate authority. Required to perform mTLS authentication.
        grpc_authentication_mode is the method of client-server authentication to use for gRPC. Default is SINGLE_USER.
        Only available on Windows.
        """
        if os.name != "nt":
            raise STKRuntimeError("STKDesktop is only available on Windows. Use STKEngine.")

        CoInitializeManager.initialize()
        if grpc_server:
            if pid is not None:
                raise STKInitializationError("Retry using either 'pid' or 'grpc_server'. Cannot initialize using both.")
            try:
                from .internal.grpcutil import GrpcClient
            except ModuleNotFoundError:
                raise STKInitializationError("gRPC use requires Python modules grpcio and protobuf.")

            if grpc_host not in ["localhost", "127.0.0.1"] and not grpc_allow_remote_host:
                raise RuntimeError("Remote host connections are not allowed. Use grpc_allow_remote_host to enable.")

            _validate_authentication_mode(grpc_authentication_mode, grpc_host)

            if grpc_authentication_mode != GrpcAuthenticationMode.MUTUAL_TLS and any([grpc_client_cert, grpc_client_key, grpc_ca]):
                raise RuntimeError("grpc_authentication_mode must be set to MUTUAL_TLS if mutual TLS parameters are provided.")

            client: GrpcClient = GrpcClient.new_client(grpc_host,
                                                       grpc_port,
                                                       grpc_timeout_sec,
                                                       grpc_max_message_size,
                                                       _get_authentication_mode_string(grpc_authentication_mode),
                                                       grpc_client_cert,
                                                       grpc_client_key,
                                                       grpc_ca)

            if client is not None:
                app_impl = client.get_stk_application_interface()
                app = STKDesktopApplication()
                app._private_init(app_impl)
                atexit.register(app._disconnect_grpc)
                return app
            else:
                raise STKInitializationError(f"Could not connect to gRPC server at {grpc_host}:{grpc_port}.")
        elif pid is None:
            clsid_aguiapplication = GUID()
            if Succeeded(OLE32Lib.CLSIDFromString("STK12.Application", clsid_aguiapplication)):
                unknown = IUnknown()
                if Succeeded(OLEAut32Lib.GetActiveObject(byref(clsid_aguiapplication), None, byref(unknown.p))):
                    unknown.take_ownership(isApplication=True)
                    app = STKDesktopApplication()
                    app._private_init(unknown)
                    return app
                else:
                    raise STKInitializationError("Failed to attach to an active STK 12 Application instance.")
        else:
            unknown = attach_to_stk_by_pid(pid)
            if unknown is not None:
                app = STKDesktopApplication()
                app._private_init(unknown)
                return app
            else:
                raise STKInitializationError("Failed to attach to STK with pid " + str(pid) + ".")

    @staticmethod
    def release_all() -> None:
        """
        Release all handles from Python to STK Desktop applications.

        Not applicable to gRPC connections.
        """
        if os.name != "nt":
            raise STKRuntimeError("STKDesktop is only available on Windows.")
        EventSubscriptionManager.unsubscribe_all()
        ObjectLifetimeManager.release_all()

    @staticmethod
    def create_thread_marshaller(stk_object:typing.Any) -> ThreadMarshaller:
        """
        Return a ThreadMarshaller instance capable of marshalling the stk_object argument to a new thread.

        Not applicable to gRPC connections.
        """
        if os.name != "nt":
            raise STKRuntimeError("STKDesktop is only available on Windows.")
        return ThreadMarshaller(stk_object)


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides graphs for Chain objects."""
import collections.abc
import typing

import matplotlib

from ansys.stk.core.stkobjects import Chain
from ansys.stk.extensions.data_analysis.graphs.graph_helpers import interval_plot, line_chart


def angle_between_line_chart(stk_object :Chain, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the angle and ranges for each three object sub-strand of a Chain, as a function of time, for each strand access interval that overlaps with the requested report time intervals.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Chain\\Angle Between.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Chain
        The STK Chain object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Angle Between').group.item("Granularity Determined").execute_elements(start_time, stop_time, step, ['Time', 'Range 1', 'Range 2', 'Angle']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'angle', 'label':'Angle', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'range 1', 'label':'Range 1', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'range 2', 'label':'Range 2', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]}]
    return line_chart([df], root, ['angle','range 1','range 2'], ['time'], axes, 'time', 'Time', 'Angle Between Data', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def bentpipe_link_cno_line_chart(stk_object :Chain, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Graph the carrier-to-noise density ratio for uplink, downlink, and composite link as a function of time for a bent pipe communications system.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Chain\\BentPipe Link - CNo.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Chain
        The STK Chain object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Link Information').execute_elements(start_time, stop_time, step, ['Time', 'C/No1', 'C/No Tot.2']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Spectral Density', 'lines': [
            {'y_name':'c/no1', 'label':'C/No1', 'use_unit':True, 'unit_squared': None, 'dimension': 'SpectralDensity'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Spectral Density', 'lines': [
            {'y_name':'c/no tot.2', 'label':'C/No Tot.2', 'use_unit':True, 'unit_squared': None, 'dimension': 'SpectralDensity'}]}]
    return line_chart([df], root, ['c/no1','c/no tot.2'], ['time'], axes, 'time', 'Time', 'Bent Pipe C/No Link Analysis', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def complete_chain_access_interval_graph(stk_object :Chain, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval plot of the time intervals for which the chain is completed.

    These intervals are computed by overlapping all the strand accesses.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Chain\\Complete Chain Access.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Chain
        The STK Chain object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Complete Access').execute_elements(start_time, stop_time, ['Start Time', 'Stop Time']).data_sets.to_pandas_dataframe()
    elements=[(('start time', 'None'),('stop time', 'None'))]
    return interval_plot([df], root, elements, [], ['start time','stop time'], 'Time', 'Complete Chain Access Times', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def individual_object_access_interval_graph(stk_object :Chain, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval graph of the time intervals for each object in a Chain for which the object participates in a strand that completes the chain.

    Each object's intervals are graphed on a separate line.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Chain\\Individual Object Access.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Chain
        The STK Chain object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Object Access').execute_elements(start_time, stop_time, ['Start Time', 'Stop Time']).data_sets.to_pandas_dataframe()
    elements=[(('start time', 'None'),('stop time', 'None'))]
    return interval_plot([df], root, elements, [], ['start time','stop time'], 'Time', 'Object Access', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def individual_strand_access_interval_graph(stk_object :Chain, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval graph of the time intervals for each strand in a Chain that completes the chain.

    Each strand's intervals are graphed on a separate line.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Chain\\Individual Strand Access.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Chain
        The STK Chain object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Strand Access').execute_elements(start_time, stop_time, ['Start Time', 'Stop Time']).data_sets.to_pandas_dataframe()
    elements=[(('start time', 'None'),('stop time', 'None'))]
    return interval_plot([df], root, elements, [], ['start time','stop time'], 'Time', 'Strand Access Times', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def number_of_accesses_line_chart(stk_object :Chain, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the number of objects participating in a strand that completes the chain at the given time, as a function of time.

    The report is only valid for Chains consisting of two objects.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\Chain\\Number Of Accesses.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.Chain
        The STK Chain object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Base Object Data').execute_elements(start_time, stop_time, step, ['Time', 'Number Of Accesses']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Value', 'lines': [
            {'y_name':'number of accesses', 'label':'Number Of Accesses', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'}]}]
    return line_chart([df], root, ['number of accesses'], ['time'], axes, 'time', 'Time', 'Number Of Accesses', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)




# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import collections
import typing

import matplotlib

from ansys.stk.core.stkobjects import STKObject
from ansys.stk.extensions.data_analysis.graphs.graph_helpers import (
    interval_pie_chart,
    interval_plot,
    line_chart,
    pie_chart,
    polar_chart,
)


def _access_duration_pie_chart(stk_object :STKObject, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.base.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Access Data').execute_elements(start_time, stop_time, ['Duration', 'Access Number']).data_sets.to_pandas_dataframe()
    return pie_chart(root, df, ['duration'], [], 'duration', 'Access Duration', 'Time', 'access number', colormap = colormap)

def _cumulative_sunlight_cumulative_pie_chart(stk_object :STKObject, start_time: typing.Any = None, stop_time : typing.Any = None, color_list: list[typing.Any] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Lighting Times').group.item('Sunlight').execute_elements(start_time, stop_time, ['Duration', 'Stop Time', 'Start Time']).data_sets.to_pandas_dataframe()
    return interval_pie_chart(root, df, ['duration'], ['start time','stop time'], 'start time', 'stop time', start_time, stop_time, 'Cumulative Sunlight', 'Time', True, color_list = color_list)

def _eclipse_times_interval_graph(stk_object :STKObject, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Eclipse Times').execute_elements(start_time, stop_time, ['Start Time', 'Stop Time']).data_sets.to_pandas_dataframe()
    elements=[(('start time', 'None'),('stop time', 'None'))]
    return interval_plot([df], root, elements, [], ['start time','stop time'], 'Time', 'Eclipse Times', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _lighting_times_interval_graph(stk_object :STKObject, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df_list=[]
    df_list.append(stk_object.data_providers.item('Lighting Times').group.item('Sunlight').execute_elements(start_time, stop_time, ['Start Time', 'Stop Time']).data_sets.to_pandas_dataframe())
    df_list.append(stk_object.data_providers.item('Lighting Times').group.item('Penumbra').execute_elements(start_time, stop_time, ['Start Time', 'Stop Time']).data_sets.to_pandas_dataframe())
    df_list.append(stk_object.data_providers.item('Lighting Times').group.item('Umbra').execute_elements(start_time, stop_time, ['Start Time', 'Stop Time']).data_sets.to_pandas_dataframe())
    elements=[(('start time', 'Sunlight'),('stop time', 'Sunlight')),(('start time', 'Penumbra'),('stop time', 'Penumbra')),(('start time', 'Umbra'),('stop time', 'Umbra'))]
    return interval_plot(df_list, root, elements, [], ['start time','stop time'], 'Time', 'Lighting Times', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _lla_position_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('LLA State').group.item('Fixed').execute_elements(start_time, stop_time, step, ['Lon', 'Time', 'Alt', 'Lat']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Latitude/Longitude', 'lines': [
            {'y_name':'lat', 'label':'Lat', 'use_unit':True, 'unit_squared': None, 'dimension': 'Latitude'},
            {'y_name':'lon', 'label':'Lon', 'use_unit':True, 'unit_squared': None, 'dimension': 'Longitude'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'alt', 'label':'Alt', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]}]
    return line_chart([df], root, ['lat','lon','alt'], ['time'], axes, 'time', 'Time', 'LLA Position', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _model_area_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Model Area').execute_elements(start_time, stop_time, step, ['Time', 'Area']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Area', 'lines': [
            {'y_name':'area', 'label':'Area', 'use_unit':True, 'unit_squared': None, 'dimension': 'Area'}]}]
    return line_chart([df], root, ['area'], ['time'], axes, 'time', 'Time', 'Model Area', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _solar_aer_line_chart_stationary(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Lighting AER').execute_elements(start_time, stop_time, step, ['Time', 'Elevation', 'Range', 'Azimuth']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'azimuth', 'label':'Azimuth', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'elevation', 'label':'Elevation', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'range', 'label':'Range', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]}]
    return line_chart([df], root, ['azimuth','elevation','range'], ['time'], axes, 'time', 'Time', 'ECF VVLH Solar AER', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _solar_az_el_polar_center_0_graph_stationary(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Lighting AER').execute_elements(start_time, stop_time, step, ['Elevation', 'Azimuth']).data_sets.to_pandas_dataframe()
    axis={'use_unit' : True, 'unit_squared': False, 'label': 'Angle', 'lines': [
        {'y_name':'elevation','x_name':'azimuth', 'label':'Azimuth', 'use_unit':True, 'unit_squared': False, 'dimension': 'Angle'}
        ]}
    return polar_chart([df], root, ['elevation','azimuth'], axis, 'Solar ECF VVLH Az-El', convert_negative_r = True, origin_0 = False, colormap=colormap)

def _solar_aer_line_chart_vehicle(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Lighting AER').group.item('ECFVVLH').execute_elements(start_time, stop_time, step, ['Time', 'Elevation', 'Range', 'Azimuth']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'azimuth', 'label':'Azimuth', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'elevation', 'label':'Elevation', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'range', 'label':'Range', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]}]
    return line_chart([df], root, ['azimuth','elevation','range'], ['time'], axes, 'time', 'Time', 'ECF VVLH Solar AER', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _solar_az_el_polar_center_0_graph_vehicle(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Lighting AER').group.item('ECFVVLH').execute_elements(start_time, stop_time, step, ['Elevation', 'Azimuth']).data_sets.to_pandas_dataframe()
    axis={'use_unit' : True, 'unit_squared': False, 'label': 'Angle', 'lines': [
        {'y_name':'elevation','x_name':'azimuth', 'label':'Azimuth', 'use_unit':True, 'unit_squared': False, 'dimension': 'Angle'}
        ]}
    return polar_chart([df], root, ['elevation','azimuth'], axis, 'Solar ECF VVLH Az-El', convert_negative_r = True, origin_0 = False, colormap=colormap)

def _solar_intensity_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Solar Intensity').execute_elements(start_time, stop_time, step, ['Time', 'Intensity']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Value', 'lines': [
            {'y_name':'intensity', 'label':'Intensity', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'}]}]
    return line_chart([df], root, ['intensity'], ['time'], axes, 'time', 'Time', 'Solar Intensity', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _solar_panel_area_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Solar Panel Area').execute_elements(start_time, stop_time, step, ['Time', 'Effective Area']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': True, 'ylog10': False, 'y2log10': False, 'label': ' Area', 'lines': [
            {'y_name':'effective area', 'label':'Effective Area', 'use_unit':True, 'unit_squared': True, 'dimension': 'SmallDistance'}]}]
    return line_chart([df], root, ['effective area'], ['time'], axes, 'time', 'Time', 'Area', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _solar_panel_power_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Solar Panel Power').execute_elements(start_time, stop_time, step, ['Time', 'Power']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Power', 'lines': [
            {'y_name':'power', 'label':'Power', 'use_unit':True, 'unit_squared': None, 'dimension': 'Power'}]}]
    return line_chart([df], root, ['power'], ['time'], axes, 'time', 'Time', 'Power', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _sun_vector_ecf_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Sun Vector').group.item('Fixed').execute_elements(start_time, stop_time, step, ['Time', 'x', 'y', 'z']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'x', 'label':'x', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'y', 'label':'y', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'z', 'label':'z', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]}]
    return line_chart([df], root, ['x','y','z'], ['time'], axes, 'time', 'Time', 'Sun Vector ECF', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _sunlight_intervals_interval_pie_chart(stk_object :STKObject, start_time: typing.Any = None, stop_time : typing.Any = None, color_list: list[typing.Any] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Lighting Times').group.item('Sunlight').execute_elements(start_time, stop_time, ['Duration', 'Stop Time', 'Start Time']).data_sets.to_pandas_dataframe()
    return interval_pie_chart(root, df, ['duration'], ['start time','stop time'], 'start time', 'stop time', start_time, stop_time, 'Sunlight Intervals', 'Time', color_list = color_list)

def _coverage_by_latitude_xy_graph(stk_object : STKObject, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    df = stk_object.data_providers.item('Coverage by Latitude').execute_elements(['Latitude', 'Percent Time Covered']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Value', 'lines': [
            {'y_name':'percent time covered', 'label':'Percent Time Covered', 'use_unit':False, 'unit_squared': None, 'dimension': 'None'}]}]
    return line_chart([df], root, ['latitude','percent time covered'], [], axes, 'latitude','Latitude', 'Coverage By Latitude', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _gap_duration_xy_graph(stk_object : STKObject, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    df = stk_object.data_providers.item('Coverage Gap Duration').execute_elements(['Duration', 'Percent Under', 'Percent Over']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : False, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Percent', 'lines': [
            {'y_name':'percent under', 'label':'% Under', 'use_unit':False, 'unit_squared': None, 'dimension': 'Percent'},
            {'y_name':'percent over', 'label':'% Over', 'use_unit':False, 'unit_squared': None, 'dimension': 'Percent'}]}]
    return line_chart([df], root, ['percent under','percent over', 'duration'], [], axes, 'duration','Duration', 'Gap Duration', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _grid_stats_over_time_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Overall Value by Time').execute_elements(start_time, stop_time, step, ['Time', 'Minimum', 'Maximum', 'Average']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'minimum', 'label':'Minimum', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'maximum', 'label':'Maximum', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'average', 'label':'Average', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['minimum','maximum','average'], ['time'], axes, 'time', 'Time', 'Grid Stats Over Time', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _value_by_latitude_xy_graph(stk_object : STKObject, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    df = stk_object.data_providers.item('Value by Latitude').execute_elements(['Maximum', 'Minimum', 'Latitude', 'Average']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'minimum', 'label':'Minimum', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'maximum', 'label':'Maximum', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'average', 'label':'Average', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['latitude','minimum','maximum','average'], [], axes, 'latitude','Latitude', 'Value By Latitude', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _value_by_longitude_xy_graph(stk_object : STKObject, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    df = stk_object.data_providers.item('Value by Longitude').execute_elements(['Minimum', 'Maximum', 'Longitude', 'Average']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'minimum', 'label':'Minimum', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'maximum', 'label':'Maximum', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'average', 'label':'Average', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['longitude','minimum','maximum','average'], [], axes, 'longitude','Longitude', 'Value By Longitude', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _lat_lon_position_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('LLA State').group.item('Fixed').execute_elements(start_time, stop_time, step, ['Lon', 'Time', 'Lat']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Latitude/Longitude', 'lines': [
            {'y_name':'lat', 'label':'Lat', 'use_unit':True, 'unit_squared': None, 'dimension': 'Latitude'},
            {'y_name':'lon', 'label':'Lon', 'use_unit':True, 'unit_squared': None, 'dimension': 'Longitude'}]}]
    return line_chart([df], root, ['lat','lon'], ['time'], axes, 'time', 'Time', 'Lat-Lon Position', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _altitude_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('LLA State').group.item('Fixed').execute_elements(start_time, stop_time, step, ['Time', 'Alt']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'alt', 'label':'Alt', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]}]
    return line_chart([df], root, ['alt'], ['time'], axes, 'time', 'Time', 'Altitude', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _beta_angle_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Beta Angle').execute_elements(start_time, stop_time, step, ['Time', 'Beta Angle']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'beta angle', 'label':'Beta Angle', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['beta angle'], ['time'], axes, 'time', 'Time', 'Beta Angle', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _euler_angles_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Euler Angles').group.item(0).execute_elements(start_time, stop_time, step, ['Time', 'C', 'A', 'B']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'a', 'label':'A', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'b', 'label':'B', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'c', 'label':'C', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['a','b','c'], ['time'], axes, 'time', 'Time', 'Euler Angles', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _fixed_position_velocity_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    data = []
    data.append(stk_object.data_providers.item('Cartesian Position').group.item('Fixed').execute_elements(start_time, stop_time, step, ['Time', 'x', 'y', 'z']).data_sets.to_pandas_dataframe())
    data.append(stk_object.data_providers.item('Cartesian Velocity').group.item('Fixed').execute_elements(start_time, stop_time, step, ['Time', 'x', 'y', 'z']).data_sets.to_pandas_dataframe())
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'x', 'label':'x', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'y', 'label':'y', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'z', 'label':'z', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Rate', 'lines': [
            {'y_name':'x', 'label':'vx', 'use_unit':True, 'unit_squared': None, 'dimension': 'Rate'},
            {'y_name':'y', 'label':'vy', 'use_unit':True, 'unit_squared': None, 'dimension': 'Rate'},
            {'y_name':'z', 'label':'vz', 'use_unit':True, 'unit_squared': None, 'dimension': 'Rate'}]}]
    return line_chart(data, root, [['x','y','z'],['x','y','z']], ['time'], axes, 'time', 'Time', 'Fixed Position & Velocity', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter, multiple_data_providers=True)

def _j2000_position_velocity_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    data = []
    data.append(stk_object.data_providers.item('Cartesian Position').group.item('J2000').execute_elements(start_time, stop_time, step, ['Time', 'x', 'y', 'z']).data_sets.to_pandas_dataframe())
    data.append(stk_object.data_providers.item('Cartesian Velocity').group.item('J2000').execute_elements(start_time, stop_time, step, ['Time', 'x', 'y', 'z']).data_sets.to_pandas_dataframe())

    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'x', 'label':'x', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'y', 'label':'y', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'z', 'label':'z', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Rate', 'lines': [
            {'y_name':'x', 'label':'vx', 'use_unit':True, 'unit_squared': None, 'dimension': 'Rate'},
            {'y_name':'y', 'label':'vy', 'use_unit':True, 'unit_squared': None, 'dimension': 'Rate'},
            {'y_name':'z', 'label':'vz', 'use_unit':True, 'unit_squared': None, 'dimension': 'Rate'}]}]
    return line_chart(data, root, [['x','y','z'],['x','y','z']], ['time'], axes, 'time', 'Time', 'J2000 Position & Velocity', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter, multiple_data_providers=True)

def _sun_vector_fixed_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Sun Vector').group.item('Fixed').execute_elements(start_time, stop_time, step, ['Time', 'x', 'y', 'z']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Distance', 'lines': [
            {'y_name':'x', 'label':'x', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'y', 'label':'y', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'},
            {'y_name':'z', 'label':'z', 'use_unit':True, 'unit_squared': None, 'dimension': 'Distance'}]}]
    return line_chart([df], root, ['x','y','z'], ['time'], axes, 'time', 'Time', 'Sun Vector Fixed', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _yaw_pitch_roll_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Attitude YPR').group.item(0).execute_elements(start_time, stop_time, step, ['Yaw', 'Time', 'Pitch', 'Roll']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'yaw', 'label':'Yaw', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'pitch', 'label':'Pitch', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'},
            {'y_name':'roll', 'label':'Roll', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['yaw','pitch','roll'], ['time'], axes, 'time', 'Time', 'YPR', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)

def _azimuth_elevation_line_chart(stk_object :STKObject, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    root = stk_object.root
    start_time = start_time or root.current_scenario.start_time
    stop_time = stop_time or root.current_scenario.stop_time
    df = stk_object.data_providers.item('Boresight AzEl').execute_elements(start_time, stop_time, step, ['Parent Rel Azimuth', 'Parent Rel Elevation', 'Time']).data_sets.to_pandas_dataframe()
    axes = [{'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'parent rel azimuth', 'label':'Azimuth', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]},
            {'use_unit' : True, 'unit_squared': None, 'ylog10': False, 'y2log10': False, 'label': 'Angle', 'lines': [
            {'y_name':'parent rel elevation', 'label':'Elevation', 'use_unit':True, 'unit_squared': None, 'dimension': 'Angle'}]}]
    return line_chart([df], root, ['parent rel azimuth','parent rel elevation'], ['time'], axes, 'time', 'Time', 'Azimuth-Elevation', colormap=colormap,  time_unit_abbreviation= time_unit_abbreviation, formatter= formatter)


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import time
import signal

from ctypes import CFUNCTYPE, cdll, c_size_t, c_int, c_void_p

from ..utilities.exceptions import STKInvalidTimerError

TIMERPROC = CFUNCTYPE(None, c_size_t)
INSTALLTIMER = CFUNCTYPE(c_size_t, c_int, TIMERPROC, c_void_p)
DELETETIMER = CFUNCTYPE(c_int, c_size_t, c_void_p)

if os.name != "nt":
    class UtilLib:
        _handle = None
        set_timer_callbacks = None
        initialize_librt_timers = None
        uninitialize_librt_timers = None
        fire_librt_timer_callbacks = None

        def initialize():
            if UtilLib._handle is None:
                UtilLib._handle = cdll.LoadLibrary("libagutil.so")
                UtilLib.set_timer_callbacks = CFUNCTYPE(None, INSTALLTIMER, DELETETIMER, c_void_p)(("AgUtSetTimerCallbacks", UtilLib._handle), ((1, "pInstallTimer"), (1, "pDeleteTimer"), (1, "pCallbackData")))
                UtilLib.initialize_librt_timers = CFUNCTYPE(None, c_int)(("AgUtInitializeLibrtTimers", UtilLib._handle), ((1, "signo"),))
                UtilLib.uninitialize_librt_timers = CFUNCTYPE(None)(("AgUtUninitializeLibrtTimers", UtilLib._handle))
                UtilLib.fire_librt_timer_callbacks = CFUNCTYPE(None)(("AgUtFireLibrtTimerCallbacks", UtilLib._handle))

class _ClockTimer(object):
    def __init__(self, id, milliseconds, TIMERPROC, callbackData):
        self.id = id
        self.interval = milliseconds/1000
        self.callback = TIMERPROC
        self.callback_data = callbackData
        self._reset()

    def _reset(self):
        self.next_proc = time.clock_gettime(time.CLOCK_REALTIME) + self.interval

    def fire(self):
        if time.clock_gettime(time.CLOCK_REALTIME) >= self.next_proc:
            self.callback(self.id)
            self._reset()

    @staticmethod
    def next_time_proc(timers:dict):
        """Return time in sec until next timer proc"""
        if len(timers) == 0:
            return 0.050
        else:
            proc_times = list()
            for timerid in timers:
                proc_times.append(timers[timerid].next_proc)
            delta_s = min(proc_times)-time.clock_gettime(time.CLOCK_REALTIME)
            if delta_s > 0:
                return delta_s
            else:
                return 0

class NullTimer(object):
    def __init__(self):
        if os.name != "nt":
            self._install_timer_cfunc = INSTALLTIMER(self.__install_timer)
            self._delete_timer_cfunc = DELETETIMER(self.__delete_timer)
            UtilLib.initialize()
            UtilLib.set_timer_callbacks(self._install_timer_cfunc, self._delete_timer_cfunc, c_void_p())
        else:
            pass

    def terminate(self):
        pass

    def __install_timer(self, milliseconds, TIMERPROC, callbackData):
        return 0

    def __delete_timer(self, timerID, callbackData):
        return 0

if os.name != "nt":
    try:
        from tkinter import Tcl
    except:
        class Tcl(object):
            def __init__(self):
                raise STKInvalidTimerError("Cannot use STKEngineTimerType.TKINTER_MAIN_LOOP nor STKEngineTimerType.INTERACTIVE_PYTHON because tkinter installation is not found.")

    class TclTimer(object):
        def __init__(self):
            self._next_id = 1
            self._timers = dict()
            self._install_timer_cfunc = INSTALLTIMER(self.__install_timer)
            self._delete_timer_cfunc = DELETETIMER(self.__delete_timer)
            UtilLib.initialize()
            UtilLib.set_timer_callbacks(self._install_timer_cfunc, self._delete_timer_cfunc, c_void_p())
            self._tcl = Tcl()
            self._tcl.after(self._next_timer_proc(), self._loop_timers)
            self._terminated = False

        def terminate(self):
            self._terminated = True
            del(self._tcl)

        def __install_timer(self, milliseconds, TIMERPROC, callbackData):
            id = self._next_id
            self._next_id = id + 1
            self._timers[id] = _ClockTimer(id, milliseconds, TIMERPROC, callbackData)
            return id

        def __delete_timer(self, timerID, callbackData):
            if timerID in self._timers:
                del(self._timers[timerID])
            return 0

        def _fire_timers(self):
            timers = self._timers.copy()
            for timerid in timers:
                timers[timerid].fire()

        def _next_timer_proc(self):
            """Return time in ms until next timer proc"""
            return int(_ClockTimer.next_time_proc(self._timers.copy())*1000)

        def _loop_timers(self):
            self._fire_timers()
            if not self._terminated:
                self._tcl.after(self._next_timer_proc(), self._loop_timers)

    class SigAlarmTimer(object):
        def __init__(self):
            self._next_id = 1
            self._timers = dict()
            self._install_timer_cfunc = INSTALLTIMER(self.__install_timer)
            self._delete_timer_cfunc = DELETETIMER(self.__delete_timer)
            UtilLib.initialize()
            UtilLib.set_timer_callbacks(self._install_timer_cfunc, self._delete_timer_cfunc, c_void_p())
            self.previous_sighandler = signal.signal(signal.SIGALRM, self._fire_timers)

        def terminate(self):
            signal.setitimer(signal.ITIMER_REAL, 0, 0)
            signal.signal(signal.SIGALRM, self.previous_sighandler)

        def __install_timer(self, milliseconds, TIMERPROC, callbackData):
            id = self._next_id
            self._next_id = id + 1
            self._timers[id] = _ClockTimer(id, milliseconds, TIMERPROC, callbackData)
            self._set_alarm_for_next_timer_proc()
            return id

        def __delete_timer(self, timerID, callbackData):
            if timerID in self._timers:
                del(self._timers[timerID])
            return 0

        def _fire_timers(self, signo, frame):
            timers = self._timers.copy()
            for timerid in timers:
                timers[timerid].fire()
            self._set_alarm_for_next_timer_proc()

        def _set_alarm_for_next_timer_proc(self):
            next_proc = _ClockTimer.next_time_proc(self._timers.copy())
            if next_proc > 0:
                signal.setitimer(signal.ITIMER_REAL, next_proc, 0)
            else:
                self._fire_timers(signal.SIGALRM, None)
                self._set_alarm_for_next_timer_proc()


    class SigRtTimer(object):
        def __init__(self, signo):
            self._next_id = 1
            self._timers = dict()
            self._signo = signo
            self.previous_sighandler = signal.signal(self._signo, self._fire_timers)
            UtilLib.initialize()
            UtilLib.initialize_librt_timers(self._signo)

        def terminate(self):
            UtilLib.uninitialize_librt_timers()
            signal.signal(self._signo, self.previous_sighandler)

        def _fire_timers(self, signo, frame):
            UtilLib.fire_librt_timer_callbacks()


# Copyright (C) 2022 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides graphs for LaunchVehicle objects."""
import collections.abc
import typing

import matplotlib

from ansys.stk.core.stkobjects import LaunchVehicle
from ansys.stk.extensions.data_analysis.graphs._common_graphs import (
    _altitude_line_chart,
    _beta_angle_line_chart,
    _cumulative_sunlight_cumulative_pie_chart,
    _eclipse_times_interval_graph,
    _euler_angles_line_chart,
    _fixed_position_velocity_line_chart,
    _j2000_position_velocity_line_chart,
    _lighting_times_interval_graph,
    _lla_position_line_chart,
    _model_area_line_chart,
    _solar_aer_line_chart_vehicle,
    _solar_az_el_polar_center_0_graph_vehicle,
    _solar_intensity_line_chart,
    _sun_vector_fixed_line_chart,
    _sunlight_intervals_interval_pie_chart,
    _yaw_pitch_roll_line_chart,
)


def altitude_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Plot the altitude of the object (i.e., magnitude of the relative position vector between the object and its detic subpoint) as a function of time

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Altitude.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _altitude_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def beta_angle_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Plot the beta angle (i.e., the signed angle of the apparent vector to the Sun) over time, relative to the orbital plane.

    The signed angle is positive when the apparent vector is in the direction of the orbit normal. The orbit normal (which is normal to the orbital plane) is parallel to the orbital angular momentum vector, which is defined as the cross-product of the inertial position and velocity vectors.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Beta Angle.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _beta_angle_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def cumulative_sunlight_cumulative_pie_chart(stk_object :LaunchVehicle, start_time: typing.Any = None, stop_time : typing.Any = None, color_list: list[typing.Any] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a pie chart showing the total duration of full sunlight within the graph's requested time interval.

    Gaps in the chart indicate the total duration of penumbra and umbra durations.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Cumulative Sunlight.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    color_list : list of typing.Any
        The colors with which to color the pie chart slices (the default is None). Must have length >= 2.

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _cumulative_sunlight_cumulative_pie_chart(stk_object, start_time, stop_time, color_list)

def eclipse_times_interval_graph(stk_object :LaunchVehicle, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval graph of the penumbra (partial lighting) and umbra (zero lighting) intervals.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Eclipse Times.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _eclipse_times_interval_graph(stk_object, start_time, stop_time, colormap, time_unit_abbreviation, formatter)

def euler_angles_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the attitude of the vehicle (i.e., the rotation between the vehicle's body axes and the vehicle' central body's inertial frame), expressed using 313 Euler angles, over time.

    Euler angles use a sequence of three rotations starting from a reference coordinate frame. The rotations are performed in succession: each rotation is relative to the frame resulting from any previous rotations. The 313 sequence uses Z, then the new X, and then finally the newest Z axis.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Euler Angles.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _euler_angles_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def fixed_position_velocity_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Plot the position and velocity of the object with respect to the object's central body, as observed from its central body's Fixed coordinate system, expressed in Cartesian components as a function of time.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Fixed Position Velocity.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _fixed_position_velocity_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def j2000_position_velocity_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Plot the position and velocity of the object with respect to the object's central body, as observed from its central body's J2000 coordinate system, expressed in Cartesian components as a function of time.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\J2000 Position Velocity.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _j2000_position_velocity_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def lighting_times_interval_graph(stk_object :LaunchVehicle, start_time: typing.Any = None, stop_time : typing.Any = None, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create an interval graph of the sunlight (full lighting) intervals, penumbra (partial lighting) intervals and umbra (zero lighting) intervals.

    Each lighting condition's intervals are plotted on separate lines.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Lighting Times.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _lighting_times_interval_graph(stk_object, start_time, stop_time, colormap, time_unit_abbreviation, formatter)

def lla_position_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Plot the position of the object, expressed in LLA elements, as a function of time.

    The coordinate system is the Fixed frame of the object's central body.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\LLA Position.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _lla_position_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def model_area_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the area of the object's 3D graphics model over time, as viewed from a given view direction, as computed by the Area Tool.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Model Area.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _model_area_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def solar_aer_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the azimuth, elevation, and range over time, describing the apparent relative position vector of the Sun with respect to Inertial VVLH axes (ECIVVLH).

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Solar AER.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _solar_aer_line_chart_vehicle(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def solar_az_el_polar_center_0_graph(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a polar plot with elevation as radius and azimuth as angle theta over time, describing the apparent relative position vector of the Sun with respect to Inertial VVLH axes (ECIVVLH).

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Solar Az-El.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _solar_az_el_polar_center_0_graph_vehicle(stk_object, start_time, stop_time, step, colormap)

def solar_intensity_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the percent of the solar disc visible over time.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Solar Intensity.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _solar_intensity_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def sun_vector_fixed_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the apparent relative position of the Sun to the object, expressed in Cartesian components, using the object's central body's Fixed coordinate system, as a function of time.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Sun Vector Fixed.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _sun_vector_fixed_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

def sunlight_intervals_interval_pie_chart(stk_object :LaunchVehicle, start_time: typing.Any = None, stop_time : typing.Any = None, color_list: list[typing.Any] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a pie chart showing each interval of full sunlight within the graph's requested time interval, separated by gaps indicating the intervals of penumbra/umbra lighting condition before and after each sunlight interval.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Sunlight Intervals.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    color_list : list of typing.Any
        The colors with which to color the pie chart slices (the default is None). Must have length >= 2.

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _sunlight_intervals_interval_pie_chart(stk_object, start_time, stop_time, color_list)

def yaw_pitch_roll_line_chart(stk_object :LaunchVehicle, start_time : typing.Any = None, stop_time :typing.Any = None, step : float = 60, colormap: matplotlib.colors.Colormap = None,  time_unit_abbreviation: str = 'UTCG', formatter: collections.abc.Callable[[float, float], str] = None) -> tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
    r"""Create a plot of the attitude of the vehicle (i.e., the rotation between the vehicle's body axes and the vehicle' central body's inertial frame), expressed using 321 YPR angles, as a function of time.

    YPR angles use a sequence of three rotations starting from a reference coordinate frame. Unlike Euler angles, the rotations are not made about axes defined by an earlier rotation: each rotation is made about the reference system's axes. The 321 sequence uses Z, then Y, and then finally the X axis.

    This graph wrapper was generated from `AGI\\STK12\\STKData\\Styles\\LaunchVehicle\\Yaw Pitch Roll.rsg`.

    Parameters
    ----------
    stk_object : ansys.stk.core.stkobjects.LaunchVehicle
        The STK LaunchVehicle object.
    start_time : typing.Any
        The start time of the calculation (the default is None, which implies using the scenario start time).
    stop_time : typing.Any
        The stop time of the calculation (the default is None, which implies using the scenario stop time).
    step_time : float
        The step time for the calculation (the default is 60 seconds).
    colormap : matplotlib.colors.Colormap
        The colormap with which to color the data (the default is None).
    time_unit_abbreviation : str
        The time unit for formatting (the default is "UTCG").
    formatter : collections.abc.Callable[[float, float], str]
        The formatter for time axes (the default is None).

    Returns
    -------
    matplotlib.figure.Figure
        The newly created figure.
    matplotlib.axes.Axes
        The newly created axes.
    """
    return _yaw_pitch_roll_line_chart(stk_object, start_time, stop_time, step, colormap, time_unit_abbreviation, formatter)

